SystemOrganization addCategory: #'Rome-Plugin'!

!CCodeGenerator methodsFor: '*Rome-Plugin' stamp: 'bf 6/27/2006 18:10'!
replaceCairoCalls
	methods do: [:m |
		m statements do: [:stmt | 
			stmt nodesDo: [:node |
				node isSend ifTrue: [node replaceCairoCalls]]]]! !

Integer subclass: #CairoConstant
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Rome-Plugin'!

!CairoConstant commentStamp: '<historical>' prior: 0!
Instances of me are created automatically for each class variable of RomePlugin starting with 'Cairo'.
When generating cCode a line like this is generated:

	#define CairoExtendRepeat CAIRO_EXTENT_REPEAT

To use Cairo constants simply create class variable with CamelCase name :-)!

!CairoConstant methodsFor: 'as yet unclassified' stamp: 'jl 9/6/2006 12:17'!
hash
	^name hash! !

!CairoConstant methodsFor: 'as yet unclassified' stamp: 'jl 9/6/2006 12:18'!
name: aConstantName
	name := (RomePlugin camelToUnderscore: aConstantName) asUppercase allButFirst! !

!CairoConstant methodsFor: 'as yet unclassified' stamp: 'jl 9/6/2006 12:20'!
printOn: aStream
	aStream nextPutAll: (name ifNil: ['undefined'])! !

!CairoConstant methodsFor: 'as yet unclassified' stamp: 'be 9/8/2006 11:01'!
printString
	^String streamContents: [ :stream | self printOn: stream ] ! !

!CairoConstant methodsFor: 'as yet unclassified' stamp: 'jl 9/6/2006 12:10'!
< aNumber
	^true! !

!CairoConstant methodsFor: 'as yet unclassified' stamp: 'jl 9/6/2006 13:28'!
= anObject
	^self == anObject
	! !

SmartSyntaxInterpreterPlugin subclass: #RomePlugin
	instanceVariableNames: 'contexts surfaceDispatch registerSurfaceFn unregisterSurfaceFn findSurfaceFn maxSurfaceID formatToDepth fontDescriptions defaultFontDescription'
	classVariableNames: 'CairoExtendRepeat CairoOperatorSource CanvasFlagFill CanvasFlagStroke CanvasFlagsIndex CanvasFontIndex CanvasFontMatrixIndex CanvasFontSizeIndex CanvasHandleIndex CanvasInstSize CanvasStackIndex CanvasStrokeColorIndex CanvasTargetIndex FormBitsIndex FormDepthIndex FormHeightIndex FormInstSize FormOffsetIndex FormWidthIndex PluginVersion TextLineBottomIndex TextLineEndIndex TextLineInternalSpaceIndex TextLineLeftIndex TextLinePaddingWidthIndex TextLineRightIndex TextLineStartIndex TextLineTopIndex'
	poolDictionaries: ''
	category: 'Rome-Plugin'!

!RomePlugin commentStamp: '<historical>' prior: 0!
Notes:

-  code generation calls #replaceCairoCalls so cairoRectangle:x:y:w:h: is generated as cairo_rectangle()


!

!RomePlugin class methodsFor: 'translation' stamp: 'bf 6/28/2006 18:53'!
camelToUnderscore: aString
	"Replace 'cairoFoo:bar:' with 'cairo_foo::' as a hack to create correct cairo calls"
	^String streamContents: [:stream |
		(aString copyUpTo: $:) do: [:c |
			c isUppercase
				ifTrue: [stream nextPut: $_; nextPut: c asLowercase]
				ifFalse: [stream nextPut: c]].
		(aString occurrencesOf: $:) timesRepeat: [
			stream nextPut: $:]]! !

!RomePlugin class methodsFor: 'translation' stamp: 'bf 7/6/2006 14:46'!
cFunc: aSelector
	"like CCG>>cFunctionNameFor: but checks for existence"
	((self whichClassIncludesSelector: aSelector)
		includesBehavior: InterpreterPlugin)
			ifFalse: [self error: 'selector not found: ', aSelector].
	^aSelector copyWithout: $:! !

!RomePlugin class methodsFor: 'translation' stamp: 'yo 6/4/2007 13:03'!
declareCVarsIn: cg
	cg addHeaderFile: '<cairo.h>'.
	cg addHeaderFile: '<cairo-ft.h>'.
	cg addHeaderFile: '<pango/pangocairo.h>'.
	cg addHeaderFile: '"SurfacePlugin.h"',
	self macroDefinitions.

	cg var: 'contexts' declareC: 'cairo_t* contexts[64]'. 

	cg var: 'fontDescriptions' declareC: 'PangoFontDescription *fontDescriptions[256]'. 
	cg var: 'defaultFontDescription' declareC: 'PangoFontDescription *defaultFontDescription'. 

	cg var: 'formatToDepth' type: 'int' array: #(32 32 8 1 16). "cairo_format_t"

	cg var: 'registerSurfaceFn' type: 'fn_ioRegisterSurface'.
	cg var: 'unregisterSurfaceFn' type: 'fn_ioUnregisterSurface'.
	cg var: 'findSurfaceFn' type: 'fn_ioFindSurface'.

	cg var: 'surfaceDispatch' declareC: 'sqSurfaceDispatch surfaceDispatch = {
  1,
  0,
  (fn_getSurfaceFormat) ', (self cFunc: #getSurfaceFormat:getW:getH:getD:getMsb:),',
  (fn_lockSurface) ', (self cFunc: #lockSurface:getPitch:x:y:w:h:), ',
  (fn_unlockSurface) ', (self cFunc: #unlockSurface:x:y:w:h:), ',
  (fn_showSurface) ', (self cFunc: #showSurface:x:y:w:h:), '
}'.

	cg replaceCairoCalls.! !

!RomePlugin class methodsFor: 'class initialization' stamp: 'bf 1/3/2008 21:35'!
initialize
	"self initialize"
	self initializePluginVersion.
	self initializeInstVarIndices.
	self initializeFlags.
	self initializeCairoConstants.
	self initializePangoConstants.

! !

!RomePlugin class methodsFor: 'class initialization' stamp: 'jl 9/6/2006 12:17'!
initializeCairoConstants
	self classPool keysDo: [:each |
		(each beginsWith: 'Cairo') ifTrue: [
			self classPool at: each put: (CairoConstant new name: each)
		]
	]! !

!RomePlugin class methodsFor: 'class initialization' stamp: 'bf 7/17/2006 18:45'!
initializeFlags

	self ensureClassPool.
	RomePluginCanvas classPool keysAndValuesDo: [:key :value |
		(key beginsWith: #Flag) ifTrue: [
			classPool at: ('Canvas', key) asSymbol put: value]].
! !

!RomePlugin class methodsFor: 'class initialization' stamp: 'bf 7/6/2006 16:53'!
initializeInstVarIndices

	self ensureClassPool.
	#(RomePluginCanvas RomePluginForm) do: [:clsName |
		| index shortName |
		shortName := clsName allButFirst: (clsName findLast: [:c | c isUppercase]) - 1.
		index := 0.
		(Smalltalk at: clsName) allInstVarNames do: [:var |
			classPool
				at: (shortName, var capitalized, 'Index') asSymbol
				put: index.
			index := index + 1].
		 classPool
			at: (shortName, 'InstSize') asSymbol
			put: index]! !

!RomePlugin class methodsFor: 'class initialization' stamp: 'yo 6/4/2007 12:07'!
initializePangoConstants

	TextLineLeftIndex := 0.
	TextLineRightIndex := 1.
	TextLineTopIndex := 2.
	TextLineBottomIndex := 3.
	TextLineStartIndex := 4.
	TextLineEndIndex := 5.
	TextLineInternalSpaceIndex := 6.
	TextLinePaddingWidthIndex := 7.
! !

!RomePlugin class methodsFor: 'class initialization' stamp: 'bf 7/31/2006 12:23'!
initializePluginVersion
	PluginVersion := self pluginVersion.
! !

!RomePlugin class methodsFor: 'translation' stamp: 'bf 7/18/2006 13:42'!
macroDefinitions
	^'
#define NUM_OF(array) (sizeof (array) / sizeof *(array))
#define lastIndex(array) (NUM_OF(array) - 1)
#define degrees(a) (a * 3.141592653589793 / 180.0)

#define log(msg) fprintf(stderr, "Squeak-Rome: " msg "\n")
#define logwith(msg, a) fprintf(stderr, "Squeak-Rome: " msg "\n", a)
#define logwithwith(msg, a, b) fprintf(stderr, "Squeak-Rome: " msg "\n", a, b)
#define logwithwithwith(msg, a, b, c) fprintf(stderr, "Squeak-Rome: " msg "\n", a, b, c)
#define logwithwithwithwith(msg, a, b, c, d) fprintf(stderr, "Squeak-Rome: " msg "\n", a, b, c, d)
#define logwithwithwithwithwith(msg, a, b, c, d, e) fprintf(stderr, "Squeak-Rome: " msg "\n", a, b, c, d, e)

#define primFail interpreterProxy->primitiveFail()
#define fail(msg) log(msg "!!"); primFail
#define failwith(msg, a) logwith(msg "!!", a); primFail
#define failwithwith(msg, a, b) logwithwith(msg "!!", a, b); primFail
'.! !

!RomePlugin class methodsFor: 'class initialization' stamp: 'bf 7/28/2006 19:36'!
moduleNameAndVersion
	^ self moduleName,
		' ', self pluginAuthor, '.', self pluginVersion asString, 
		' ', Date today asString! !

!RomePlugin class methodsFor: 'class initialization' stamp: 'bf 7/28/2006 19:35'!
pluginAuthor
	MCWorkingCopy managersForCategory: self category do: [:wc |
		^(wc ancestors first name copyAfterLast: $-)
			copyUpToLast: $.].
	self error: 'cannot get author initials'! !

!RomePlugin class methodsFor: 'class initialization' stamp: 'bf 4/23/2008 16:22'!
pluginVersion
	Smalltalk at: #MCWorkingCopy ifPresent: [:mcWorkingCopy |
		mcWorkingCopy managersForClass: self do: [:wc |
			| s v |
			v := ((wc ancestors ifEmpty: [ChangeSet current] ifNotEmptyDo: [:a | a first])
				name copyAfterLast: $.) asInteger.
			v notNil ifTrue: [
				^(wc modified and: [InMidstOfFileinNotification signal not])
					ifTrue: [v+1]
					ifFalse: [v]]]].

	InMidstOfFileinNotification signal
		ifTrue: [^nil].

	[PluginVersion isNil] whileTrue: [
		PluginVersion := (FillInTheBlank request: 'RomePlugin version number
(should match MC package version)') asInteger].

	^PluginVersion! !

!RomePlugin class methodsFor: 'translation' stamp: 'bf 2/29/2008 09:37'!
translateInDirectory: directory doInlining: inlineFlag
	| exports |
	exports := super translateInDirectory: directory doInlining: inlineFlag.
	Smalltalk at: #RomePluginCanvas ifPresent: [:canvasClass |
		(canvasClass expectedPluginVersion < self pluginVersion
			and: [Utilities authorInitialsPerSe = 'bf'])
				ifTrue: [canvasClass expectedPluginVersion: self pluginVersion]].
	^exports
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 12/12/2006 00:41'!
addAlignment: attrArrayOop into: pangoAttrList
	| start end alignment pangoAlignment attrArray |
	self var: 'attrArray' type: 'sqInt *'.
	self var: 'pangoAttrList' type: 'PangoAttrList *'.
	self var: 'pangoAlignment' type: 'int'.
	self var: 'pangoAttr' type: 'PangoAttribute *'.

	attrArray := interpreterProxy firstIndexableField: attrArrayOop.
	start := interpreterProxy integerValueOf: (attrArray at: 1).
	end := interpreterProxy integerValueOf: (attrArray at: 2).
	alignment := interpreterProxy integerValueOf: (attrArray at: 3).
	alignment = 0 ifTrue: [pangoAlignment := self cCode: 'PANGO_ALIGN_LEFT'].
	alignment = 1 ifTrue: [pangoAlignment := self cCode: 'PANGO_ALIGN_RIGHT'].
	alignment = 2 ifTrue: [pangoAlignment := self cCode: 'PANGO_ALIGN_CENTER'].

	alignment = 3 ifTrue: ["self pangoLayoutSetJustify"].
"
	pangoAttr := self pango"! !

!RomePlugin methodsFor: 'patterns' stamp: 'bf 7/28/2006 15:00'!
addColorStopTo: pattern offset: intOffset rgb: rgb alpha: alpha
	| r g b |
	self returnTypeC: 'void'.
	self var: 'pattern' type: 'cairo_pattern_t*'.

	r := rgb >> 20 bitAnd: 1023.
	g := rgb >> 10 bitAnd: 1023.
	b := rgb >> 0 bitAnd: 1023.
	alpha = 255
		ifTrue: [self cairoPatternAddColorStopRgb: pattern
			offset: intOffset / 65536.0 r: r/1023.0 g: g /1023.0 b: b/1023.0]
		ifFalse: [self cairoPatternAddColorStopRgba: pattern
			offset: intOffset / 65536.0 r: r/1023.0 g: g /1023.0 b: b/1023.0 a: alpha/255.0].
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 12/12/2006 16:40'!
addColor: attrArrayOop into: pangoAttrList
	| start end c alpha r g b pangoAttr attrArray |
	self var: 'attrArray' type: 'sqInt *'.
	self var: 'pangoAttrList' type: 'PangoAttrList *'.
	self var: 'pangoAttr' type: 'PangoAttribute *'.
	self var: 'c' type: 'unsigned int'.

	attrArray := interpreterProxy firstIndexableField: attrArrayOop.
	start := interpreterProxy integerValueOf: (attrArray at: 1).
	end := interpreterProxy integerValueOf: (attrArray at: 2).
	c := interpreterProxy positive32BitValueOf: (attrArray at: 3).
	self log: 'color: %u' with: c.
	alpha := (c &16rFF000000) >> 24.
	r := (c bitAnd: 16r00FF0000)>>16.
	g := (c bitAnd: 16r0000FF00)>>8.
	b := (c bitAnd: 16r000000FF).
	r = 0 ifFalse: [r := r * 257].
	g = 0 ifFalse: [g := g * 257].
	b = 0 ifFalse: [b := b * 257].

	self log: '%d %d %d %d' with: alpha with: r with: g with: b.
	pangoAttr := self pangoAttrForegroundNew: r g: g b: b.
	self cCode: 'pangoAttr->start_index = start'.
	self cCode: 'pangoAttr->end_index = end'.
	self pangoAttrListChange: pangoAttrList attr: pangoAttr.
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/11/2007 14:44'!
addDefaultInto: pangoAttrList
	| pangoAttr lang |
	self var: 'pangoAttrList' type: 'PangoAttrList *'.
	self var: 'pangoAttr' type: 'PangoAttribute *'.
	self var: 'timesDesc' type: 'PangoFontDescription *'.
	self var: 'lang' type: 'PangoLanguage *'.

	pangoAttr := self pangoAttrForegroundNew: 0 g: 0 b: 1.
	self cCode: 'pangoAttr->start_index = 0'.
	self cCode: 'pangoAttr->end_index = 0x7fffffff'.
	self pangoAttrListInsert: pangoAttrList attr: pangoAttr.

"	pangoAttr := self pangoAttrBackgroundNew: 0 g: 0 b: 0.
	self cCode: 'pangoAttr->start_index = 0'.
	self cCode: 'pangoAttr->end_index = 0x7fffffff'.
	self pangoAttrListInsert: pangoAttrList attr: pangoAttr."

	pangoAttr := self pangoAttrFontDescNew: defaultFontDescription.
	self cCode: 'pangoAttr->start_index = 0'.
	self cCode: 'pangoAttr->end_index = 0x7fffffff'.
	self pangoAttrListInsert: pangoAttrList attr: pangoAttr.

	lang := self pangoLanguageFromString: 'en-US'.
	pangoAttr := self pangoAttrLanguageNew: lang.
	self cCode: 'pangoAttr->start_index = 0'.
	self cCode: 'pangoAttr->end_index = 0x7fffffff'.
	self pangoAttrListInsert: pangoAttrList attr: pangoAttr.

! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 12/12/2006 13:45'!
addEmphasis: attrArrayOop into: pangoAttrList
	| start end c pangoAttr attrArray |
	self var: 'attrArray' type: 'sqInt *'.
	self var: 'pangoAttrList' type: 'PangoAttrList *'.
	self var: 'pangoAttr' type: 'PangoAttribute *'.

	attrArray := interpreterProxy firstIndexableField: attrArrayOop.
	start := interpreterProxy integerValueOf: (attrArray at: 1).
	end := interpreterProxy integerValueOf: (attrArray at: 2).
	c := interpreterProxy integerValueOf: (attrArray at: 3).

	(c bitAnd: 1) ifTrue: [
		pangoAttr := self pangoAttrWeightNew: (self cCode: 'PANGO_WEIGHT_BOLD').     
		self cCode: 'pangoAttr->start_index = start'.
		self cCode: 'pangoAttr->end_index = end'.
		self pangoAttrListChange: pangoAttrList attr: pangoAttr.
	].

	(c bitAnd: 2) ifTrue: [
		pangoAttr := self pangoAttrStyleNew: (self cCode: 'PANGO_STYLE_OBLIQUE').     
		self cCode: 'pangoAttr->start_index = start'.
		self cCode: 'pangoAttr->end_index = end'.
		self pangoAttrListChange: pangoAttrList attr: pangoAttr.
	].
	(c bitAnd: 4) ifTrue: [
		pangoAttr := self pangoAttrUnderlineNew: (self cCode: 'PANGO_UNDERLINE_SINGLE').     
		self cCode: 'pangoAttr->start_index = start'.
		self cCode: 'pangoAttr->end_index = end'.
		self pangoAttrListChange: pangoAttrList attr: pangoAttr.
	].
	(c bitAnd: 8) ifTrue: [
		pangoAttr := self pangoAttrStretchNew: (self cCode: 'PANGO_STRETCH_CONDENSED').     
		self cCode: 'pangoAttr->start_index = start'.
		self cCode: 'pangoAttr->end_index = end'.
		self pangoAttrListChange: pangoAttrList attr: pangoAttr.
	].
	(c bitAnd: 16) ifTrue: [
		pangoAttr := self pangoAttrStrikethroughNew: true.     
		self cCode: 'pangoAttr->start_index = start'.
		self cCode: 'pangoAttr->end_index = end'.
		self pangoAttrListChange: pangoAttrList attr: pangoAttr.
	].
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/9/2007 21:50'!
addFont: attrArrayOop into: pangoAttrList
	| start end pangoAttr attrArray desc fontDescIndex |
	self var: 'attrArray' type: 'sqInt *'.
	self var: 'pangoAttrList' type: 'PangoAttrList *'.
	self var: 'pangoAttr' type: 'PangoAttribute *'.
	self var: 'desc' type: 'PangoFontDescription *'.

	attrArray := interpreterProxy firstIndexableField: attrArrayOop.
	start := interpreterProxy integerValueOf: (attrArray at: 1).
	end := interpreterProxy integerValueOf: (attrArray at: 2).
	fontDescIndex := interpreterProxy fetchInteger: 5 ofObject: (attrArray at: 4).

	fontDescIndex < 0 ifTrue: [^ nil].
	fontDescIndex > 255 ifTrue: [^ nil].
	desc := fontDescriptions at: fontDescIndex.
	desc = nil ifTrue: [^ nil].
	pangoAttr := self pangoAttrFontDescNew: desc.
	self cCode: 'pangoAttr->start_index = start'.
	self cCode: 'pangoAttr->end_index = end'.
	self pangoAttrListChange: pangoAttrList attr: pangoAttr.
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/11/2007 10:38'!
addLanguage: attrArrayOop into: pangoAttrList
	| start end lang pangoAttr attrArray cLang pangoLang |
	self var: 'attrArray' type: 'sqInt *'.
	self var: 'pangoAttrList' type: 'PangoAttrList *'.
	self var: 'pangoAttr' type: 'PangoAttribute *'.
	self var: 'pangoLang' type: 'PangoLanguage *'.
	self var: 'cLang' type: 'char *'.

	attrArray := interpreterProxy firstIndexableField: attrArrayOop.
	start := interpreterProxy integerValueOf: (attrArray at: 1).
	end := interpreterProxy integerValueOf: (attrArray at: 2).
	lang := interpreterProxy integerValueOf: (attrArray at: 3).

	lang = 0 ifTrue: [
		cLang := 'en-US'.
	].
	lang = 5 ifTrue: [
		cLang := 'ja-JP'.
	].
	lang = 6 ifTrue: [
		cLang := 'zh-CN'.
	].
	lang = 7 ifTrue: [
		cLang := 'ko-KR'.
	].
	lang = 9 ifTrue: [
		cLang := 'zh-TW'.
	].
	lang = 13 ifTrue: [
		cLang := 'el-EL'.
	].
	pangoLang := self pangoLanguageFromString: cLang.
	pangoAttr := self pangoAttrLanguageNew: pangoLang.
	self cCode: 'pangoAttr->start_index = start'.
	self cCode: 'pangoAttr->end_index = end'.
	self pangoAttrListChange: pangoAttrList attr: pangoAttr.
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/11/2007 17:30'!
addSelectionAt: rect pixel: c with: context
	| alpha r g b |
	self var: 'context' type: 'cairo_t*'.
	self var: 'rect' type: 'PangoRectangle *'.
	self var: 'c' type: 'unsigned int'.
	
	alpha := (c &16rFF000000) >> 24.
	r := (c bitAnd: 16r00FF0000)>>16.
	g := (c bitAnd: 16r0000FF00)>>8.
	b := (c bitAnd: 16r000000FF).
	(alpha = 0 and: [r = 0 and: [g = 0 and: [b = 0]]]) ifTrue: [^ nil].
	r = 0 ifFalse: [r := r * 257].
	g = 0 ifFalse: [g := g * 257].
	b = 0 ifFalse: [b := b * 257].

	self cairoSave: context.
	self cairoSetSourceRgb: context r: r g: g b: b.
	self cairoNewPath: context.
	self	cairoMoveTo: context x: (self cCode: 'PANGO_PIXELS(rect->x)') y: (self cCode: 'PANGO_PIXELS(rect->y)').
	self	cairoLineTo: context x: (self cCode: 'PANGO_PIXELS(rect->x)') y: (self cCode: 'PANGO_PIXELS(rect->y+rect->height)').
	self cairoStroke: context.
	self cairoRestore: context.
	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/11/2007 17:30'!
addSelectionFrom: start to: end pixel: c into: pangoAttrList
	| pangoAttr alpha r g b |
	self var: 'pangoAttrList' type: 'PangoAttrList *'.
	self var: 'pangoAttr' type: 'PangoAttribute *'.
	self var: 'c' type: 'unsigned int'.

	alpha := (c &16rFF000000) >> 24.
	r := (c bitAnd: 16r00FF0000)>>16.
	g := (c bitAnd: 16r0000FF00)>>8.
	b := (c bitAnd: 16r000000FF).
	(alpha = 0 and: [r = 0 and: [g = 0 and: [b = 0]]]) ifTrue: [^ nil].
	r = 0 ifFalse: [r := r * 257].
	g = 0 ifFalse: [g := g * 257].
	b = 0 ifFalse: [b := b * 257].

 	pangoAttr := self pangoAttrBackgroundNew: r with: g with: b.
	self cCode: 'pangoAttr->start_index = start'.
	self cCode: 'pangoAttr->end_index = end'.
	self pangoAttrListChange: pangoAttrList attr: pangoAttr.
! !

!RomePlugin methodsFor: 'contexts' stamp: 'bf 7/24/2006 15:40'!
contextFrom: canvasOop
	"Get contexts[canvasOop's handle]"

	| context contextIndex targetOop |
	self inline: false.
	self returnTypeC: 'cairo_t*'.
	self var: 'context' type: 'cairo_t*'.

	(interpreterProxy slotSizeOf: canvasOop) < CanvasInstSize	ifTrue: [
		self fail: 'canvas oop invalid'. ^nil].
	contextIndex := interpreterProxy fetchInteger: CanvasHandleIndex ofObject: canvasOop.
	interpreterProxy failed ifTrue: [
		contextIndex := interpreterProxy fetchPointer: CanvasHandleIndex ofObject: canvasOop.
		contextIndex = interpreterProxy nilObject "fail silently"
			ifFalse: [self fail: 'canvas handle not an integer']. 
		^nil].

	(contextIndex < 0 or: [contextIndex > contexts lastIndex]) ifTrue: [
		self fail: 'canvas handle %i out of bounds' with: contextIndex. ^nil].

	context := contexts at: contextIndex.
	context isNil ifTrue: [
		self fail: 'canvas handle %i invalid' with: contextIndex. ^nil].

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].

	"Check surface matches, too"
	targetOop := interpreterProxy fetchPointer: CanvasTargetIndex ofObject: canvasOop.
	(self surfaceFrom: targetOop) = (self cairoGetTarget: context)
		ifFalse: [self fail: 'canvas handle %i does not match target' with: contextIndex. ^nil].

	^context! !

!RomePlugin methodsFor: 'contexts' stamp: 'bf 7/25/2006 18:30'!
createContextFor: canvasOop

	| targetOop targetID targetSurface context contextIndex i  |
	self inline: false.

	self var: 'targetSurface' type: 'cairo_surface_t*'.
	self var: 'context' type: 'cairo_t*'.

	(interpreterProxy slotSizeOf: canvasOop) < CanvasInstSize ifTrue: [
		self fail: 'canvas oop invalid'. ^nil].
	targetOop := interpreterProxy fetchPointer: CanvasTargetIndex ofObject: canvasOop.

	(interpreterProxy slotSizeOf: targetOop) <= FormBitsIndex ifTrue: [
		self fail: 'target oop invalid'. ^nil].
	targetID := interpreterProxy fetchInteger: FormBitsIndex ofObject: targetOop.
	interpreterProxy failed	ifTrue: [
		self fail: 'target handle not an integer'. ^nil].

	targetSurface := self findSurface: targetID.
	(targetSurface notNil and: [(self cairoSurfaceStatus: targetSurface) = 0]) ifFalse: [
		self fail: 'target surface invalid'. ^nil].

	contextIndex := -1.
	i := 0.
	[i <= contexts lastIndex] whileTrue: [
		(contexts at: i) isNil	ifTrue: [
			contextIndex := i.
			i := contexts lastIndex].
		i := i + 1].
	contextIndex < 0 ifTrue: [
		self fail: 'too many canvases'. ^nil].

	context := self cairoCreate: targetSurface.

	contexts at: contextIndex put: context.
	self debugCode: [self log: 'created context %i' with: contextIndex].

	^contextIndex! !

!RomePlugin methodsFor: 'surfaces' stamp: 'bf 7/7/2006 19:45'!
createSurfaceFor: formOop
	"create a surface, register it in SurfacePlugin and answer its surface plugin ID"
	| format surface width height depth bits status |
	self var: 'format' type: 'cairo_format_t'.
	self var: 'surface' type: 'cairo_surface_t*'.

	(interpreterProxy slotSizeOf: formOop) < FormInstSize	ifTrue: [
		self fail: 'form oop invalid'. ^-1].
	width := interpreterProxy fetchInteger: FormWidthIndex ofObject: formOop.	
	height := interpreterProxy fetchInteger: FormHeightIndex ofObject: formOop.	
	depth := interpreterProxy fetchInteger: FormDepthIndex ofObject: formOop.
	interpreterProxy failed ifTrue: [
		self fail: 'form fields are not integers'. ^-1].

	self cCode: 'switch(depth) {
		case 32: format = CAIRO_FORMAT_ARGB32; break;
		case 24: format = CAIRO_FORMAT_RGB24; break;
		case 16: format = CAIRO_FORMAT_RGB16_565; break;
		case  8: format = CAIRO_FORMAT_A8; break;
		case  1: format = CAIRO_FORMAT_A1; break;
		default: format = -1;
	}' inSmalltalk: [format := depth].

	(width <= 0 or: [height <= 0 or: [format < 0]]) ifTrue: [
		self fail: 'form fields out of range'. ^-1].

	bits := interpreterProxy fetchPointer: FormBitsIndex ofObject: formOop.	
	bits = interpreterProxy nilObject	ifFalse: [
		self fail: 'form handle not nil'. ^-1].
	
	surface := self cairoImageSurfaceCreate: format width: width height: height.
	status := self cairoSurfaceStatus: surface.
	status = 0 ifFalse: [
		self fail: 'failed to create surface - %s' with: (self cairoStatusToString: status).
		self cairoSurfaceDestroy: surface.
		^-1].

	^self registerSurface: surface.
! !

!RomePlugin methodsFor: 'contexts' stamp: 'bf 7/25/2006 18:30'!
destroyContextFor: canvasOop

	| contextIndex context |
	self inline: false.
	self var: 'context' type: 'cairo_t*'.

	(interpreterProxy slotSizeOf: canvasOop) < CanvasInstSize	ifTrue: [
		self fail: 'canvas oop invalid'. ^nil].
	contextIndex := interpreterProxy fetchInteger: CanvasHandleIndex ofObject: canvasOop.
	interpreterProxy failed ifTrue: [
		self fail: 'canvas handle not an integer'. ^nil].

	(contextIndex < 0 or: [contextIndex > contexts lastIndex]) ifTrue: [
		self fail: 'canvas handle %i out of bounds' with: contextIndex. ^nil].

	context := contexts at: contextIndex.
	context isNil ifTrue: [
		self fail: 'canvas handle %i invalid' with: contextIndex. ^nil].

	self debugCode: [self log: 'destroyed context %i' with: contextIndex].

	self cairoDestroy: context.

	contexts at: contextIndex put: nil.
	^nil! !

!RomePlugin methodsFor: 'surfaces' stamp: 'bf 7/7/2006 20:04'!
destroySurface: surfaceID
	"fetch surface from surfaceID, destroy it and unregister from SurfacePlugin"
	| surface |
	self var: 'surface' type: 'cairo_surface_t*'.

	surface := self findSurface: surfaceID.
	surface isNil ifTrue: [
		self fail: 'could not find surface %i' with: surfaceID. ^nil].

	self cairoSurfaceDestroy: surface.
	
	self unregisterSurface: surfaceID.
	^nil! !

!RomePlugin methodsFor: 'contexts' stamp: 'bf 7/17/2006 19:36'!
fillOrStroke: context from: canvasOop
	"fill or stroke depending on canvasOop's flags"

	| canvasFlags fill stroke rgb |
	self returnTypeC: 'void'.
	self var: 'context' type: 'cairo_t*'.

	canvasFlags := interpreterProxy fetchInteger: CanvasFlagsIndex ofObject: canvasOop.
	interpreterProxy failed
		ifTrue: [self fail: 'canvas flags not an integer']
		ifFalse: [
			fill := canvasFlags bitAnd: CanvasFlagFill.
			stroke := canvasFlags bitAnd: CanvasFlagStroke.
			fill ~= 0 ifTrue: [
				stroke ~= 0
					ifTrue: [self cairoFillPreserve: context]
					ifFalse: [self cairoFill: context]].
			stroke ~= 0
				ifTrue: [
					rgb := interpreterProxy fetchInteger: CanvasStrokeColorIndex ofObject: canvasOop.
					self cairoSave: context.
					self setSource: context rgb: rgb alpha: stroke. 
					self cairoStroke: context.
					self cairoRestore: context]].! !

!RomePlugin methodsFor: 'surfaces' stamp: 'bf 7/7/2006 19:54'!
findSurface: surfaceID
	"Answer surface handle for surfaceID"
	| surfaceHandle |
	self returnTypeC: 'cairo_surface_t*'.

	findSurfaceFn isNil ifTrue: [
		self loadSurfacePlugin ifFalse: [^nil]].

	(self cCode: '(*findSurfaceFn)(surfaceID, &surfaceDispatch, &surfaceHandle)' inSmalltalk: [surfaceHandle := surfaceID])
		ifFalse: [^nil].

	^self cCoerce: surfaceHandle to: 'cairo_surface_t*'.

	! !

!RomePlugin methodsFor: 'surface-callbacks' stamp: 'bf 7/7/2006 18:07'!
getSurfaceFormat: surfaceHandle getW: wReturn getH: hReturn getD: dReturn getMsb: mReturn
	| width height depth msb |
	self var: 'surfaceHandle' type: 'cairo_surface_t *'.
	self var: 'wReturn' type: 'int*'.
	self var: 'hReturn' type: 'int*'.
	self var: 'dReturn' type: 'int*'.
	self var: 'mReturn' type: 'int*'.

	self debugCode: [self log: 'get surface format %p' with: surfaceHandle].

	width := self cairoImageSurfaceGetWidth: surfaceHandle.
	height := self cairoImageSurfaceGetHeight: surfaceHandle.
	depth := formatToDepth at: (self cairoImageSurfaceGetFormat: surfaceHandle).
	msb := true.

	self cCode: '*wReturn = width' inSmalltalk: [wReturn value: width].
	self cCode: '*hReturn = height' inSmalltalk: [hReturn value: height].
	self cCode: '*dReturn = depth' inSmalltalk: [dReturn value: depth].
	self cCode: '*mReturn = msb' inSmalltalk: [mReturn value: msb].

	^true! !

!RomePlugin methodsFor: 'initialize' stamp: 'yo 6/4/2007 13:04'!
initialiseModule
	self export: true.

	0 to: contexts lastIndex do: [:i |
		contexts at: i put: nil].
	0 to: fontDescriptions lastIndex do: [:i |
		fontDescriptions at: i put: nil].
	
	defaultFontDescription := self pangoFontDescriptionFromString: 'Times New Roman 10'.

	registerSurfaceFn := nil.
	unregisterSurfaceFn := nil.
	findSurfaceFn := nil.
	^true! !

!RomePlugin methodsFor: 'surfaces' stamp: 'bf 7/10/2006 11:48'!
loadSurfacePlugin
	"Load the surface support plugin"
	| found |
	registerSurfaceFn := self cCoerce: (interpreterProxy ioLoadFunction: 'ioRegisterSurface' From: 'SurfacePlugin')
		to: 'fn_ioRegisterSurface'.
	unregisterSurfaceFn :=  self cCoerce: (interpreterProxy ioLoadFunction: 'ioUnregisterSurface' From: 'SurfacePlugin')
		to: 'fn_ioUnregisterSurface'.
	findSurfaceFn :=  self cCoerce: (interpreterProxy ioLoadFunction: 'ioFindSurface' From: 'SurfacePlugin')
		to: 'fn_ioFindSurface'.
	found := registerSurfaceFn notNil and: [unregisterSurfaceFn notNil and: [findSurfaceFn notNil]].
	found ifFalse: [self fail: 'could not load SurfacePlugin'].

	maxSurfaceID := -1.
	^found! !

!RomePlugin methodsFor: 'surface-callbacks' stamp: 'bf 7/7/2006 18:08'!
lockSurface: surfaceHandle getPitch: pitchReturn x: x y: y w: w h: h
	| data pitch |
	self returnTypeC: 'unsigned char*'.
	self var: 'surfaceHandle' type: 'cairo_surface_t*'.
	self var: 'pitchReturn' type: 'int*'.
	self var: 'data' type: 'unsigned char*'.

	self debugCode: [
		self log: 'lock surface %p %i %i %i %i' with: surfaceHandle
			with: x with: y with: w with: h].

	self cairoSurfaceFlush: surfaceHandle.
	data := self cairoImageSurfaceGetData: surfaceHandle.
	pitch := self cairoImageSurfaceGetStride: surfaceHandle.

	self cCode: '*pitchReturn = pitch' inSmalltalk: [pitchReturn value: pitch].
	^data! !

!RomePlugin methodsFor: 'initialize' stamp: 'bf 7/6/2006 18:10'!
moduleUnloaded: aModuleName
	"The module with the given name was just unloaded.
	Make sure we have no dangling references."
	self export: true.
	self var: #aModuleName type: 'char *'.
	(aModuleName strcmp: 'SurfacePlugin') = 0 ifTrue:[
		registerSurfaceFn := nil.
		unregisterSurfaceFn := nil.
		findSurfaceFn := nil].
	^true! !

!RomePlugin methodsFor: 'contexts' stamp: 'bf 7/25/2006 17:47'!
polyPath: context from: pointsOop
	| points pointsCount |
	self returnTypeC: 'void'.
	self var: 'context' type: 'cairo_t*'.
	self var: 'points' type: 'float*'.

	points := interpreterProxy arrayValueOf: pointsOop.
	pointsCount := interpreterProxy slotSizeOf: pointsOop.
	interpreterProxy failed ifFalse: [
		pointsCount >= 2 ifTrue:[
			self cairoMoveTo: context x: (points at: 0) y: (points at: 1).
			2 to: pointsCount-1 by: 2 do: [:i |
				self cairoLineTo: context x: (points at: i) y: (points at: i+1)].
		].
	].
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/5/2007 14:49'!
primitiveBlockInString: aString at: utf8Index x: x y: y squeakAttributeArray: sqAttrArray width: w height: h withWrap: withWrap into: charData
	| canvasOop context layout attrList index trailing pos origin corner atEnd |
	self var: 'context' type: 'cairo_t*'.
	self var: 'layout' type: 'PangoLayout*'.
	self var: 'line' type: 'PangoLayoutLine*'.
	self var: 'attrList' type: 'PangoAttrList*'.
	self var: 'run' type: 'PangoLayoutRun*'.
	self var: 'pos' type: 'PangoRectangle'.

	self cCode: '' inSmalltalk: [index := index. trailing := trailing. pos := pos].

	canvasOop := self primitive: 'primitivePangoBlockAtIndex' parameters: #(String SmallInteger SmallInteger SmallInteger Object SmallInteger SmallInteger Boolean Object).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	layout := self pangoCairoCreateLayout: context.
	attrList := self pangoAttrListNew.
	self translateSqAttrsToPangoAttrs: sqAttrArray into: attrList.
	self pangoLayoutSetText: layout string: aString length: (self cCode: 'strlen(aString)').
	self pangoLayoutSetAttributes: layout attrs: attrList.
	self pangoLayoutSetWidth: layout width: (w * (self cCode: 'PANGO_SCALE')).
	withWrap ifTrue: [
		self pangoLayoutSetWrap: layout wrapMode: (self cCode: 'PANGO_WRAP_WORD_CHAR')].

	self cairoTranslate: context x: x y: y.
	self pangoCairoUpdateLayout: context layout: layout.

	(utf8Index = ((self cCode: 'strlen(aString)') + 1)) ifTrue: [
		atEnd := true.
		self pangoLayoutIndexToPos: layout with: utf8Index - 1 with: (self cCode: '&pos').
		self cCode: 'pos.x = pos.x + pos.width'.
	] ifFalse: [
		atEnd := false.
		self pangoLayoutIndexToPos: layout with: utf8Index with: (self cCode: '&pos').
	].

	self pangoAttrListUnref: attrList.
	self cCode: 'g_object_unref(layout)'.
	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].

	origin := interpreterProxy makePointwithxValue: (self cCode: 'PANGO_PIXELS(pos.x)') + x yValue: (self cCode: 'PANGO_PIXELS(pos.y)') + y.
	corner := interpreterProxy makePointwithxValue: (self cCode: 'PANGO_PIXELS(pos.x+pos.width)') + x yValue: (self cCode: 'PANGO_PIXELS(pos.y+pos.height)') + y.
	interpreterProxy storePointer: 0 ofObject: charData withValue: origin.
	interpreterProxy storePointer: 1 ofObject: charData withValue: corner.
	index := (self sqCharCountIn: aString from: 0 to: utf8Index).
	atEnd ifFalse: [
		index := index + 1.
	].
	interpreterProxy storePointer: 2 ofObject: charData withValue: (interpreterProxy integerObjectOf: index).
	^ index asOop: SmallInteger.
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'jl 9/6/2006 12:21'!
primitiveClear
	| canvasOop context canvasFlags fill op |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveClear'.
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	canvasFlags := interpreterProxy fetchInteger: CanvasFlagsIndex ofObject: canvasOop.
	interpreterProxy failed
		ifTrue: [self fail: 'canvas flags not an integer'. ^nil].

	"if no fill was selected, select transparent color"
	fill := canvasFlags bitAnd: CanvasFlagFill.
	fill ifFalse: [self cairoSetSourceRgba: context r: 0.0 g: 0.0 b: 0.0 a: 0.0].

	op := context cairoGetOperator.
	context cairoSetOperator: CairoOperatorSource.
	self cairoPaint: context.
	context cairoSetOperator: op.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/24/2006 16:25'!
primitiveClipRectangleLeft: left right: right top: top bottom: bottom
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveClipRectangleLeftRightTopBottom' parameters: #(Float Float Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoNewPath: context.
	self	cairoRectangle: context x: left y: top w: right - left h: bottom - top.
	self cairoClip: context.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-canvas' stamp: 'bf 7/3/2006 15:30'!
primitiveClose
	| canvasOop |
	canvasOop := self primitive: 'primitiveClose'.

	self destroyContextFor: canvasOop.

	interpreterProxy failed ifFalse: [
		interpreterProxy storePointer: CanvasHandleIndex
			ofObject: canvasOop
			withValue: interpreterProxy nilObject.
	].

! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/5/2007 23:26'!
primitiveComposeString: aString x: x y: y squeakAttributeArray: sqAttrArray width: w height: h withWrap: withWrap into: lines
	| canvasOop context layout attrList lineCount lineIter baseline line start ink logical sqStart sqEnd totalY prevBaseline addition linesSize lastLine newW next |
	self var: 'context' type: 'cairo_t*'.
	self var: 'layout' type: 'PangoLayout*'.
	self var: 'line' type: 'PangoLayoutLine*'.
	self var: 'attrList' type: 'PangoAttrList*'.
	self var: 'lineIter' type: 'PangoLayoutIter*'.
	self var: 'ink' type: 'PangoRectangle'.
	self var: 'logical' type: 'PangoRectangle'.
	self var: 'run' type: 'PangoLayoutRun*'.
	self var: 'aString' type: 'unsigned char*'.
	
	self cCode: '' inSmalltalk: [addition := addition. ink := ink. logical := logical. prevBaseline := prevBaseline].

	canvasOop := self primitive: 'primitivePangoComposeString' parameters: #(String SmallInteger SmallInteger Object SmallInteger SmallInteger Boolean Object).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	layout := self pangoCairoCreateLayout: context.
	attrList := self pangoAttrListNew.
	self translateSqAttrsToPangoAttrs: sqAttrArray into: attrList.
	self pangoLayoutSetText: layout string: aString length: (self cCode: 'strlen(aString)').
	self pangoLayoutSetAttributes: layout attrs: attrList.
	self pangoLayoutSetWidth: layout width: (w * (self cCode: 'PANGO_SCALE')).
	withWrap ifTrue: [
		self pangoLayoutSetWrap: layout wrapMode: (self cCode: 'PANGO_WRAP_WORD_CHAR')].

	self cairoTranslate: context x: x y: y.
	self pangoCairoUpdateLayout: context layout: layout.
	lineCount := self pangoLayoutGetLineCount: layout.
	lineIter := self pangoLayoutGetIter: layout.
	baseline := self pangoLayoutIterGetBaseline: lineIter.
	prevBaseline := 0.
	
	self cairoTranslate: context x: 0 y: (self cCode: 'PANGO_PIXELS(baseline)').
	self log: 'lineCount %d %d' with: lineCount with: (self cCode: 'strlen(aString)').

	sqEnd := 0.
	totalY := 0.
	linesSize := interpreterProxy stSizeOf: lines.
	lastLine := (lineCount min: linesSize) - 1.
	0 to: lastLine do: [:lineIndex |
		line := self pangoLayoutIterGetLineReadonly: lineIter.
		self pangoLayoutLineGetExtents: line ink: (self cCode: '&ink') logical: (self cCode: '&logical').
		sqStart := sqEnd + 1.
		start := self cCode: 'line->start_index'.
		sqEnd := sqStart + (self sqCharCountIn: aString from: start to: (start + (self cCode: 'line->length'))) - 1.
		self log: 'sqStart, sqEnd: %d %d' with: sqStart with: sqEnd.
		((start + (self cCode: 'line->length')) < (interpreterProxy stSizeOf: (self cCoerce: aString to: 'int')) and: [
			(aString at: (start + (self cCode: 'line->length'))) = 13]) ifTrue: [
				sqEnd := sqEnd + 1.
		].
		self log: 'sqEnd: %d' with: sqEnd.
		self log: 'logical %d %d %d %d' with: (self cCode: 'logical.x') with: (self cCode: 'logical.y') with: (self cCode: 'logical.width') with: (self cCode: 'logical.height').
		interpreterProxy storeInteger: TextLineStartIndex ofObject: (interpreterProxy fetchPointer: lineIndex ofObject: lines) withValue: sqStart.
		interpreterProxy storeInteger: TextLineEndIndex ofObject: (interpreterProxy fetchPointer: lineIndex ofObject: lines) withValue: sqEnd.
		interpreterProxy storeInteger: TextLineLeftIndex ofObject: (interpreterProxy fetchPointer: lineIndex ofObject: lines) withValue: (self cCode: 'PANGO_PIXELS(logical.x)') + x.
		interpreterProxy storeInteger: TextLineRightIndex ofObject: (interpreterProxy fetchPointer: lineIndex ofObject: lines) withValue: (self cCode: 'PANGO_PIXELS(logical.x+logical.width)') + x.
		interpreterProxy storeInteger: TextLineTopIndex ofObject: (interpreterProxy fetchPointer: lineIndex ofObject: lines) withValue: totalY + y.
		interpreterProxy storeInteger: TextLineInternalSpaceIndex ofObject: (interpreterProxy fetchPointer: lineIndex ofObject: lines) withValue: (interpreterProxy integerObjectOf: 0).
		interpreterProxy storeInteger: TextLinePaddingWidthIndex ofObject: (interpreterProxy fetchPointer: lineIndex ofObject: lines) withValue: (interpreterProxy integerObjectOf: 0).

		next := self pangoLayoutIterNextLine: lineIter.

		next
			ifTrue: [addition := (self cCode: 'logical.height') + (self pangoLayoutGetSpacing: layout)]
			ifFalse: [addition := (self cCode: 'logical.height')].

		totalY := totalY + (self cCode: 'PANGO_PIXELS(addition)').
		interpreterProxy storeInteger: TextLineBottomIndex ofObject: (interpreterProxy fetchPointer: lineIndex ofObject: lines) withValue: totalY + y.
		
	].

	self pangoLayoutGetExtents: layout with: (self cCode: '&ink') with: (self cCode: '&logical').
	newW := self cCode: 'PANGO_PIXELS(logical.width)'.

	self pangoLayoutIterFree: lineIter.
	self pangoAttrListUnref: attrList.        
	self cCode: 'g_object_unref(layout)'.
	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
	^ newW asOop: SmallInteger.
! !

!RomePlugin methodsFor: 'primitives-form' stamp: 'bf 7/7/2006 11:19'!
primitiveCreateFormHandle
	| formOop surfaceID |
	formOop := self primitive: 'primitiveCreateFormHandle'.

	surfaceID := self createSurfaceFor: formOop.
	surfaceID < 0 ifTrue: [interpreterProxy primitiveFail].

	interpreterProxy failed
		ifFalse: [^surfaceID asSmallIntegerObj].
! !

!RomePlugin methodsFor: 'primitives-form' stamp: 'bf 7/7/2006 11:28'!
primitiveDestroyFormHandle: surfaceID
	self primitive: 'primitiveDestroyFormHandle' parameters: #(SmallInteger).

	interpreterProxy failed
		ifFalse: [self destroySurface: surfaceID].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/24/2006 17:54'!
primitiveDrawArc: radius x: x y: y from: angle1 to: angle2
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveDrawArcRadiusXYFromTo' parameters: #(Float Float Float Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoNewPath: context.
	radius > 0.0
		ifTrue: [self cairoArc: context x: x y: y r: radius from: angle1 to: angle2]
		ifFalse: [self cairoArcNegative: context x: x y: y r: -1.0 * radius from: angle1 to: angle2].
	self	fillOrStroke: context from: canvasOop.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/24/2006 17:54'!
primitiveDrawCurveFromX: x0 y: y0 viaX: x1 y: y1 andX: x2 y: y2 toX: x3 y: y3
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveDrawCurveFromXYviaXYandXYtoXY' parameters: #(Float Float Float Float Float Float Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoNewPath: context.
	self cairoMoveTo: context x: x0 y: y0.
	self cairoCurveTo: context viaX: x1 y: y1 andX: x2 y: y2 toX: x3 y: y3.
	self	fillOrStroke: context from: canvasOop.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/24/2006 18:23'!
primitiveDrawCurveFromX: x0 y: y0 viaX: x1 y: y1 toX: x2 y: y2
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveDrawCurveFromXYviaXYtoXY' parameters: #(Float Float Float Float Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoNewPath: context.
	self cairoMoveTo: context x: x0 y: y0.
	self cairoCurveTo: context 
		viaX: 2.0 * x1 + x0 / 3.0 y: 2.0 * y1 + y0 / 3.0
		andX: 2.0 * x1 + x2 / 3.0 y: 2.0 * y1 + y2 / 3.0
		toX: x2 y: y2.
	self	fillOrStroke: context from: canvasOop.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/24/2006 18:02'!
primitiveDrawGeneralBezierShape: shapeOop
	| canvasOop context  contourOop shapeCount segmentCount seg |
	self var: 'context' type: 'cairo_t*'.
	self var: 'seg' type: 'short*'.
	canvasOop := self primitive: 'primitiveDrawGeneralBezierShape' parameters: #(Oop).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	shapeCount := interpreterProxy slotSizeOf: shapeOop.
	self cairoNewPath: context.

	0 to: shapeCount - 1 do: [:i | 
		contourOop := interpreterProxy fetchPointer: i ofObject: shapeOop.
		(interpreterProxy is: contourOop MemberOf: 'ShortPointArray')
			ifFalse: [self fail: 'bezier contour %i is no ShortPointArray' with: i+1. ^nil].
		segmentCount := (interpreterProxy slotSizeOf: contourOop) // 3.
		seg := interpreterProxy arrayValueOf: contourOop.
	
		0 to: segmentCount - 1 do: [:j |
			| xFrom yFrom xVia yVia xTo yTo |
			xFrom := seg at: 0.
			yFrom := seg at: 1.
			xVia := seg at: 2.
			yVia := seg at: 3.
			xTo := seg at: 4.
			yTo := seg at: 5.
			
			j = 0 ifTrue: [self cairoMoveTo: context x: xFrom y: yFrom].

			(xFrom = xVia and: [yFrom = yVia])
				ifTrue: [self cairoLineTo: context x: xTo y: yTo]
				ifFalse: [
					"increase degree to cubic: via1 = 1/3 from + 2/3 via, via2 = 2/3 via + 1/3 to"
					self cairoCurveTo: context
						viaX: xFrom + (2.0 * xVia) // 3.0 y: yFrom + (2.0 * yVia) // 3.0
						andX: 2.0 * xVia + xTo // 3.0 y: 2.0 * yVia + yTo // 3.0
						toX: xTo y: yTo].
			seg := seg + 6.
		].

		self cairoClosePath: context.
	].

	self	fillOrStroke: context from: canvasOop.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/29/2006 00:15'!
primitiveDrawImage: formOop destLeft: dstL right: dstR top: dstT bottom: dstB srcLeft: srcL right: srcR top: srcT bottom: srcB
	| canvasOop context srcSurface |
	self var: 'context' type: 'cairo_t*'.
	self var: 'srcSurface' type: 'cairo_surface_t*'.

	canvasOop := self primitive: 'primitiveDrawImageSrcLRTBDestLRTB' parameters: #(Oop Float Float Float Float Float Float Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	srcSurface := self surfaceFrom: formOop.
	(srcSurface notNil and: [(self cairoSurfaceStatus: srcSurface) = 0]) ifFalse: [
		self fail: 'source surface invalid'. ^nil].

	(srcR ~= srcL and: [srcT ~= srcB]) ifTrue: [
		self cairoSave: context.
		self cairoTranslate: context x: dstL y: dstT.
		self cairoScale: context x: (dstR - dstL) / (srcR - srcL) y: (dstB - dstT) / (srcB - srcT).
		self cairoNewPath: context.
		self cairoRectangle: context x: 0.0 y: 0.0 w: srcR - srcL h: srcB - srcT.
		self cairoSetSourceSurface: context surface: srcSurface x: 0.0-srcL y: 0.0-srcT.
		self cairoFill: context.
		self cairoRestore: context].

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/24/2006 16:46'!
primitiveDrawLineFromX: fromX y: fromY toX: toX y: toY
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveDrawLineFromXYtoXY' parameters: #(Float Float Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoNewPath: context.
	self	cairoMoveTo: context x: fromX y: fromY.
	self	cairoLineTo: context x: toX y: toY.
	self	stroke: context from: canvasOop.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/24/2006 19:20'!
primitiveDrawOvalLeft: left right: right top: top bottom: bottom
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveDrawOvalLeftRightTopBottom' parameters: #(Float Float Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	(right ~= left and: [top ~= bottom]) ifTrue: [
		self cairoSave: context.
		self cairoNewPath: context.
		self	cairoTranslate: context x: left + right / 2.0 y: top + bottom / 2.0.
		self cairoScale: context x: right - left / 2.0 y: bottom - top / 2.0.
		self cairoArc: context x: 0.0 y: 0.0 r: 1.0 from: 0.0 to: 360.0 degrees.
		"restore transform before stroke so line width is not affected"
		self cairoRestore: context.
		self	fillOrStroke: context from: canvasOop.
	].

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/25/2006 17:44'!
primitiveDrawPolygon: pointsOop
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveDrawPolygon' parameters: #(Oop).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoNewPath: context.
	self polyPath: context from: pointsOop.
	self cairoClosePath: context.
	self	fillOrStroke: context from: canvasOop.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/25/2006 17:45'!
primitiveDrawPolyline: pointsOop
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveDrawPolyline' parameters: #(Oop).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoNewPath: context.
	self polyPath: context from: pointsOop.
	self	fillOrStroke: context from: canvasOop.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/24/2006 15:40'!
primitiveDrawRectangleLeft: left right: right top: top bottom: bottom
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveDrawRectangleLeftRightTopBottom' parameters: #(Float Float Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoNewPath: context.
	self	cairoRectangle: context x: left y: top w: right - left h: bottom - top.

	self	fillOrStroke: context from: canvasOop.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/31/2006 14:14'!
primitiveDrawRoundRectLeft: x1 right: x4 top: y1 bottom: y4 radius: radius corners: cornerFlags
	"	x1	x2		x3	x4
	y1	+	/	-	\	+
	y2	/	.		.	\
		|				|
	y3	\	.		.	/
	y4	+	\	-	/	+"
	| canvasOop context rx ry r x2 x3 y2 y3 |
	self var: 'context' type: 'cairo_t*'.
	self var: 'rx' type: 'double'.
	self var: 'ry' type: 'double'.
	self var: 'r' type: 'double'.
	self var: 'x2' type: 'double'.
	self var: 'x3' type: 'double'.
	self var: 'y2' type: 'double'.
	self var: 'y3' type: 'double'.

	canvasOop := self primitive: 'primitiveDrawRoundRectLeftRightTopBottomRadiusCorner'
		parameters: #(Float Float Float Float Float SmallInteger).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	rx := x4 - x1 / 2.0.
	ry := y4 - y1 / 2.0.
	r := (radius min: rx) min: ry.
	x2 := x1 + r.
	x3 := x4 - r.
	y2 := y1 + r.
	y3 := y4 - r.

	self cairoNewPath: context.

	(cornerFlags anyMask: 2r0001) ifTrue: [
		self cairoArc: context x: x2 y: y2 r: r from: 180 degrees to: 270 degrees.
	] ifFalse: [
		self cairoMoveTo: context x: x1 y: y1.
	].

	(cornerFlags anyMask: 2r1000) ifTrue: [
		self cairoArc: context x: x3 y: y2 r: r from: 270 degrees to: 360 degrees.
	] ifFalse: [
		self cairoLineTo: context x: x4 y: y1.
	].

	(cornerFlags anyMask: 2r0100) ifTrue: [
		self cairoArc: context x: x3 y: y3 r: r from: 0 degrees to: 90 degrees.
	] ifFalse: [
		self cairoLineTo: context x: x4 y: y4.
	].

	(cornerFlags anyMask: 2r0010) ifTrue: [
		self cairoArc: context x: x2 y: y3 r: r from: 90 degrees to: 180 degrees.
	] ifFalse: [
		self cairoLineTo: context x: x1 y: y4.
	].

	self cairoClosePath: context.

	self	fillOrStroke: context from: canvasOop.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-text' stamp: 'bf 7/24/2006 15:41'!
primitiveDrawZeroTerminatedUtf8String: aString x: x y: y
	"draw a string including its outline if selected"
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveDrawZeroTerminatedUtf8StringXY' parameters: #(String Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].
	
	self cairoMoveTo: context x: x y: y.
	context cairoTextPath: aString.
	self	fillOrStroke: context from: canvasOop.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-attributes' stamp: 'jl 9/6/2006 11:57'!
primitiveFillBitmapOriginX: ox y: oy directionX: dx y: dy normalX: nx y: ny repeat: aBoolean image: formOop
	| canvasOop context pattern m mdata surface |
	self var: 'context' type: 'cairo_t*'.
	self var: 'pattern' type: 'cairo_pattern_t*'.
	self var: 'surface' type: 'cairo_surface_t*'.
	self var: 'mdata' type: 'cairo_matrix_t'.
	self var: 'm' type: 'cairo_matrix_t*'.
	m := (self cCode: '&mdata' inSmalltalk: [mdata := MatrixTransform2x3 new. mdata]).

	canvasOop := self primitive: 'primitiveFillBitmapOriginXYdirectionXYnormalXYRepeatImage'
		parameters: #(Float Float Float Float Float Float Boolean Oop).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	surface := self surfaceFrom: formOop.
	(surface notNil and: [(self cairoSurfaceStatus: surface) = 0]) ifFalse: [
		self fail: 'surface invalid'. ^nil].

	pattern := self cairoPatternCreateForSurface: surface.
	self cairoMatrixInit: m
		xx: dx	yx: dy
		xy: nx	yy: ny
		x0: ox	y0: oy.
	m cairoMatrixInvert.
	pattern cairoPatternSetMatrix: m.
	aBoolean ifTrue: [
		pattern cairoPatternSetExtend: CairoExtendRepeat. 
	].
	pattern cairoPatternStatus ifTrue: [
				self fail: 'cairo error: %s' with: pattern cairoPatternStatus cairoStatusToString. ^nil].

	context cairoSetSource: pattern.
	pattern cairoPatternDestroy.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-attributes' stamp: 'bf 7/24/2006 15:40'!
primitiveFillColor: rgb alpha: alpha
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveFillColorAlpha' parameters: #(SmallInteger SmallInteger).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self setSource: context rgb: rgb alpha: alpha.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-attributes' stamp: 'bf 7/28/2006 14:55'!
primitiveFillLinearGradientOriginX: ox y: oy directionX: dx y: dy colorStops: colorStopsOop
	| canvasOop context pattern colorStops colorStopsCount |
	self var: 'context' type: 'cairo_t*'.
	self var: 'pattern' type: 'cairo_pattern_t*'.
	self var: 'colorStops' type: 'int*'.

	canvasOop := self primitive: 'primitiveFillLinearOriginXYdirectionXYcolorStops'
		parameters: #(Float Float Float Float Oop).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	colorStops := interpreterProxy arrayValueOf: colorStopsOop.
	colorStopsCount := interpreterProxy slotSizeOf: colorStopsOop.
	interpreterProxy failed ifTrue: [
		self fail: 'bad colorStops array'. ^nil].

	pattern := self cairoPatternCreateLinear: ox y: oy toX: ox+dx y: oy+dy.
	pattern cairoPatternStatus ifTrue: [
		self fail: 'cairo error: %s' with: pattern cairoPatternStatus cairoStatusToString. ^nil].

	0 to: colorStopsCount-1 by: 3 do: [:i |
		self addColorStopTo: pattern
			offset: (colorStops at: i)
			rgb: (colorStops at: i+1)
			alpha: (colorStops at: i+2)].
	context cairoSetSource: pattern.
	pattern cairoPatternDestroy.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-attributes' stamp: 'jl 9/6/2006 10:59'!
primitiveFillRadialGradientOriginX: ox y: oy directionX: dx y: dy normalX: nx y: ny colorStops: colorStopsOop
	| canvasOop context pattern colorStops colorStopsCount m mdata |
	self var: 'context' type: 'cairo_t*'.
	self var: 'pattern' type: 'cairo_pattern_t*'.
	self var: 'colorStops' type: 'int*'.
	self var: 'mdata' type: 'cairo_matrix_t'.
	self var: 'm' type: 'cairo_matrix_t*'.
	m := (self cCode: '&mdata' inSmalltalk: [mdata := MatrixTransform2x3 new. mdata]).

	canvasOop := self primitive: 'primitiveFillRadialOriginXYdirectionXYnormalXYcolorStops'
		parameters: #(Float Float Float Float Float Float Oop).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	colorStops := interpreterProxy arrayValueOf: colorStopsOop.
	colorStopsCount := interpreterProxy slotSizeOf: colorStopsOop.
	interpreterProxy failed ifTrue: [
		self fail: 'bad colorStops array'. ^nil].

	pattern := self cairoPatternCreateRadial: 0.0 y: 0.0 r: 0.0 x: 0.0 y: 0.0 r: 1.0.
	self cairoMatrixInit: m
		xx: dx	yx: dy
		xy: nx	yy: ny
		x0: ox	y0: oy.
	m cairoMatrixInvert.
	pattern cairoPatternSetMatrix: m.
	pattern cairoPatternStatus ifTrue: [
				self fail: 'cairo error: %s' with: pattern cairoPatternStatus cairoStatusToString. ^nil].

	0 to: colorStopsCount-1 by: 3 do: [:i |
		self addColorStopTo: pattern
			offset: (colorStops at: i)
			rgb: (colorStops at: i+1)
			alpha: (colorStops at: i+2)].
	context cairoSetSource: pattern.
	pattern cairoPatternDestroy.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-text' stamp: 'bf 7/24/2006 15:41'!
primitiveFontFace: faceHandle
	| canvasOop context ftFace crFace |
	self var: 'context' type: 'cairo_t*'.
	self var: 'ftFace' type: 'FT_Face'.
	self var: 'crFace' type: 'cairo_font_face_t*'.

	canvasOop := self primitive: 'primitiveFontFace' parameters: #(ByteArray).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].
	
	ftFace := self cCode: '*(void**)faceHandle'.
	crFace := self cairoFtFontFaceCreateForFtFace: ftFace loadFlags: (self cCode: 'FT_LOAD_NO_HINTING').
	context cairoSetFontFace: crFace.
	crFace cairoFontFaceDestroy.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-text' stamp: 'bf 7/24/2006 15:41'!
primitiveFontSize: size
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveFontSize' parameters: #(Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].
	
	context cairoSetFontSize: size.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-attributes' stamp: 'bf 7/24/2006 15:41'!
primitiveGetLineWidth
	| canvasOop context width |
	self var: 'context' type: 'cairo_t*'.
	self var: 'width' type: 'double'.

	canvasOop := self primitive: 'primitiveGetLineWidth'.
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	width := self cairoGetLineWidth: context.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].

	^width asFloatObj! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/5/2007 21:38'!
primitiveGetPangoDescriptionFamilyName: familyNameOop size: ptSize
	| familyNameSize familyName cFamilyName newInd ind desc isAbsolute size pangoStyle |
	self var: 'context' type: 'cairo_t*'.
	self var: 'cFamilyName' declareC: 'char cFamilyName[128]'.
	self var: 'familyName' declareC: 'char *familyName'.
	self var: 'desc' type: 'PangoFontDescription *'.
	
	self cCode: '' inSmalltalk: [cFamilyName := cFamilyName].

	self primitive: 'primitivePangoFontDescriptionIndex' parameters: #(Object "(String)" SmallInteger).
	interpreterProxy failed ifTrue: [^ nil].

	familyNameSize := interpreterProxy stSizeOf: familyNameOop.
	familyNameSize > 127 ifTrue: [^ nil].

	(ptSize < 0 or: [ptSize > 1000000]) ifTrue: [^ interpreterProxy integerObjectOf: -1].

	familyName := interpreterProxy firstIndexableField: familyNameOop.
	0 to: familyNameSize -1 do: [:i |
		cFamilyName at: i put: (familyName at: i).
	].
	cFamilyName at: familyNameSize put: 0.

	newInd := -1.
	ind := 0.
	desc := nil.
	[newInd = -1] whileTrue: [
		desc := fontDescriptions at: ind.
		desc = nil ifTrue: [newInd := ind]
		ifFalse: [
			isAbsolute := self pangoFontDescriptionGetSizeIsAbsolute: desc.
			size := self pangoFontDescriptionGetSize: desc.
			isAbsolute ifTrue: [
				size := size.
			].
			((self cCode: 'strcmp(cFamilyName, pango_font_description_get_family(desc))') = 0
				and: [size = (ptSize * (self cCode: 'PANGO_SCALE'))]) ifTrue: [
					newInd := ind.
			].
		].
		ind := ind + 1.
		ind > 255 ifTrue: [newInd := 256].
	].
	newInd = 256 ifTrue: [^ interpreterProxy integerObjectOf: -1].
	desc ifNotNil: [^ interpreterProxy integerObjectOf: newInd].

	desc := self pangoFontDescriptionNew.
	self pangoFontDescriptionSetFamily: desc family: cFamilyName.
	pangoStyle := self cCode: 'PANGO_STYLE_NORMAL'.
	self pangoFontDescriptionSetStyle: desc style: pangoStyle.
	self pangoFontDescriptionSetSize: desc style: ptSize * (self cCode: 'PANGO_SCALE').
	fontDescriptions at: newInd put: desc.
	^ interpreterProxy integerObjectOf: newInd.
! !

!RomePlugin methodsFor: 'primitives-transform' stamp: 'bf 7/24/2006 15:41'!
primitiveGetTransform: aTransform
	| canvasOop context m |
	self var: 'context' type: 'cairo_t*'.
	self var: 'm' type: 'cairo_matrix_t'.

	canvasOop := self primitive: 'primitiveGetTransform' parameters: #(FloatArray).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoGetMatrix: context matrix: (self cCode: '&m' inSmalltalk: [m := m]).

	aTransform at: 0 put: (self cCode: 'm.xx').
	aTransform at: 1 put: (self cCode: 'm.xy').
	aTransform at: 2 put: (self cCode: 'm.x0').
	aTransform at: 3 put: (self cCode: 'm.yx').
	aTransform at: 4 put: (self cCode: 'm.yy').
	aTransform at: 5 put: (self cCode: 'm.y0').

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/5/2007 23:38'!
primitiveIndexInString: aString x: x y: y squeakAttributeArray: sqAttrArray width: w height: h withWrap: withWrap into: charData
	| canvasOop context layout attrList inside index trailing pos origin corner lineIndex line xPos |
	self var: 'context' type: 'cairo_t*'.
	self var: 'layout' type: 'PangoLayout*'.
	self var: 'attrList' type: 'PangoAttrList*'.
	self var: 'run' type: 'PangoLayoutRun*'.
	self var: 'pos' type: 'PangoRectangle'.
	self var: 'line' type: 'PangoLayoutLine*'.
	
	self cCode: '' inSmalltalk: [index := index. trailing := trailing. pos := pos. xPos := xPos. lineIndex := lineIndex].

	canvasOop := self primitive: 'primitivePangoIndexAtPoint' parameters: #(String SmallInteger SmallInteger Object SmallInteger SmallInteger Boolean Object).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	layout := self pangoCairoCreateLayout: context.
	attrList := self pangoAttrListNew.
	self translateSqAttrsToPangoAttrs: sqAttrArray into: attrList.
	self pangoLayoutSetText: layout string: aString length: (self cCode: 'strlen(aString)').
	self pangoLayoutSetAttributes: layout attrs: attrList.
	self pangoLayoutSetWidth: layout width: (w * (self cCode: 'PANGO_SCALE')).
	withWrap ifTrue: [
		self pangoLayoutSetWrap: layout wrapMode: (self cCode: 'PANGO_WRAP_WORD_CHAR')].

	self cairoTranslate: context x: x y: y.
	self pangoCairoUpdateLayout: context layout: layout.

	"lineIter := self pangoLayoutGetIter: layout.
	baseline := self pangoLayoutIterGetBaseline: lineIter.
	self pangoLayoutIterFree: lineIter."
	inside := self pangoLayoutXyToIndex: layout with: (x * (self cCode: 'PANGO_SCALE')) with: (y * (self cCode: 'PANGO_SCALE')) with: (self cCode: '&index') with: (self cCode: '&trailing').
	self pangoLayoutIndexToPos: layout with: index with: (self cCode: '&pos').

	line := nil.
	inside ifFalse: [
		self pangoLayoutIndexToLineX: layout with: index with: trailing with: (self cCode: '&lineIndex') with: (self cCode: '&xPos').
		line := self pangoLayoutGetLineReadonly: layout with: lineIndex.
		line = nil ifTrue: [
			index := 0.
		] ifFalse: [
			trailing ifFalse: [
				index := (self cCode: 'line->start_index').
			].
		].
	].

	line = nil ifFalse: [self pangoLayoutLineUnref: line].
	self pangoAttrListUnref: attrList.
	self cCode: 'g_object_unref(layout)'.
	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
	
	origin := interpreterProxy makePointwithxValue: (self cCode: 'PANGO_PIXELS(pos.x)') yValue: (self cCode: 'PANGO_PIXELS(pos.y)').
	corner := interpreterProxy makePointwithxValue: (self cCode: 'PANGO_PIXELS(pos.x+pos.width)') yValue: (self cCode: 'PANGO_PIXELS(pos.y+pos.height)').
	interpreterProxy storePointer: 0 ofObject: charData withValue: origin.
	interpreterProxy storePointer: 1 ofObject: charData withValue: corner.
	index := (self sqCharCountIn: aString from: 0 to: (index + trailing)) + 1.
	interpreterProxy storePointer: 2 ofObject: charData withValue: (interpreterProxy integerObjectOf: index).
	
	^ index asOop: SmallInteger.
! !

!RomePlugin methodsFor: 'primitives-canvas' stamp: 'bf 7/3/2006 15:30'!
primitiveOpen
	| canvasOop contextIndex handleOop |
	canvasOop := self primitive: 'primitiveOpen'.

	contextIndex := self createContextFor: canvasOop.

	interpreterProxy failed ifFalse: [
		handleOop := contextIndex asSmallIntegerObj.
		interpreterProxy storePointer: CanvasHandleIndex ofObject: canvasOop withValue: handleOop.
	].
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/5/2007 19:39'!
primitivePangoIsAvailable
	| canvasOop layout context |
	self var: 'context' type: 'cairo_t*'.
	self var: 'layout' type: 'PangoLayout*'.
	
	canvasOop := self primitive: 'primitivePangoIsAvailable' parameters: #().
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^ nil].

	layout := self pangoCairoCreateLayout: context.
	layout = nil ifTrue: [
		^ 2 asOop: SmallInteger
	].
	self cCode: 'g_object_unref(layout)'.
	^ true asOop: Boolean
! !

!RomePlugin methodsFor: 'primitives' stamp: 'bf 6/26/2006 15:37'!
primitivePluginVersion
	self primitive: 'primitivePluginVersion'.
	^ PluginVersion asSmallIntegerObj! !

!RomePlugin methodsFor: 'primitives-canvas' stamp: 'bf 7/24/2006 15:41'!
primitiveRestoreState
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveRestoreState'.
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoRestore: context.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-transform' stamp: 'bf 7/24/2006 15:42'!
primitiveRotateBy: angle
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveRotateBy' parameters: #(Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoRotate: context angle: angle.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-canvas' stamp: 'bf 7/24/2006 15:42'!
primitiveSaveState
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveSaveState'.
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoSave: context.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-transform' stamp: 'bf 7/24/2006 15:42'!
primitiveScaleByX: x y: y
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveScaleBy' parameters: #(Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoScale: context x: x y: y.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-attributes' stamp: 'bf 7/24/2006 15:42'!
primitiveSetLineWidth: width
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveSetLineWidth' parameters: #(Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoSetLineWidth: context width: width.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-transform' stamp: 'bf 7/24/2006 15:42'!
primitiveSetTransform: aTransform
	| canvasOop context m |
	self var: 'context' type: 'cairo_t*'.
	self var: 'm' type: 'cairo_matrix_t'.

	canvasOop := self primitive: 'primitiveSetTransform' parameters: #(FloatArray).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].
	
	self cairoMatrixInit: (self cCode: '&m' inSmalltalk: [m := aTransform])
		xx: (aTransform at: 0)
		yx: (aTransform at: 3)
		xy: (aTransform at: 1)
		yy: (aTransform at: 4)
		x0: (aTransform at: 2)
		y0: (aTransform at: 5).

	self cairoSetMatrix: context matrix: (self cCode: '&m' inSmalltalk: [m]).

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/5/2007 23:28'!
primitiveShowString: aString x: x y: y squeakAttributeArray: sqAttrArray width: w height: h withWrap: withWrap selectionStart: selStart selectionEnd: selEnd selectionColorPixel: cPixel
	| canvasOop context layout attrList lineCount lineIter baseline line ink logical prevBaseline indexToX startRect newW cursorColor |
	self var: 'context' type: 'cairo_t*'.
	self var: 'layout' type: 'PangoLayout*'.
	self var: 'line' type: 'PangoLayoutLine*'.
	self var: 'attrList' type: 'PangoAttrList*'.
	self var: 'lineIter' type: 'PangoLayoutIter*'.
	self var: 'ink' type: 'PangoRectangle'.
	self var: 'logical' type: 'PangoRectangle'.
	self var: 'startRect' type: 'PangoRectangle'.
	self var: 'run' type: 'PangoLayoutRun*'.
	self var: 'cursorColor' type: 'unsigned int'.
	
	self cCode: '' inSmalltalk: [prevBaseline := prevBaseline. ink := ink. logical := logical. indexToX := indexToX. startRect := startRect].

	canvasOop := self primitive: 'primitivePangoShowString' parameters: #(String SmallInteger SmallInteger Object SmallInteger SmallInteger Boolean SmallInteger SmallInteger Object).
	context := self contextFrom: canvasOop.
	
	cursorColor := interpreterProxy positive32BitValueOf: cPixel.
	interpreterProxy failed ifTrue: [^nil].
	
	layout := self pangoCairoCreateLayout: context.
	attrList := self pangoAttrListNew.
	self translateSqAttrsToPangoAttrs: sqAttrArray into: attrList.
	
	selStart = selEnd ifFalse: [
		(selStart >= 0 and: [selEnd >= 0]) ifTrue: [
			self addSelectionFrom: selStart to: selEnd pixel: cursorColor into: attrList.
		].
	].
	self pangoLayoutSetText: layout string: aString length: (self cCode: 'strlen(aString)').
	self pangoLayoutSetAttributes: layout attrs: attrList.
	self pangoLayoutSetWidth: layout width: (w * (self cCode: 'PANGO_SCALE')).
	withWrap ifTrue: [
		self pangoLayoutSetWrap: layout wrapMode: (self cCode: 'PANGO_WRAP_WORD_CHAR')].

	self cairoTranslate: context x: x y: y.
	self pangoCairoUpdateLayout: context layout: layout.
	lineCount := self pangoLayoutGetLineCount: layout.
"	lineIter := self pangoLayoutGetIter: layout.
	baseline := self pangoLayoutIterGetBaseline: lineIter.
	prevBaseline := 0.
"
	true ifTrue: [
		self pangoCairoShowLayout: context layout: layout.
	] ifFalse: [
"	self cairoTranslate: context x: 0 y: (self cCode: 'PANGO_PIXELS(baseline)').
	0 to: lineCount - 1 do: [:lineIndex |
		line := self pangoLayoutIterGetLineReadonly: lineIter.
		
		self pangoLayoutLineGetExtents: line ink: (self cCode: '&ink') logical: (self cCode: '&logical').
		self cairoTranslate: context x: (self cCode: '-(PANGO_PIXELS(ink.x))') y: 0.
		self pangoCairoShowLayoutLine: context line: line.
		self pangoLayoutIterNextLine: lineIter.
		prevBaseline := baseline.
		baseline := self pangoLayoutIterGetBaseline: lineIter.
		self cairoTranslate: context x: (self cCode: 'PANGO_PIXELS(ink.x)')
			y: (self cCode: 'PANGO_PIXELS(baseline-prevBaseline)').
	].
"	].

	selStart = selEnd ifTrue: [
		(selStart >= 0 and: [selEnd >= 0]) ifTrue: [
			self pangoLayoutIndexToPos: layout with: selStart with: (self cCode: '&startRect').
			self addSelectionAt: (self cCode: '&startRect') pixel: cursorColor with: context.
		]
	].

	self pangoLayoutGetExtents: layout with: (self cCode: '&ink') with: (self cCode: '&logical').
	newW := self cCode: 'PANGO_PIXELS(logical.width)'.
	"self pangoLayoutIterFree: lineIter."
	self pangoAttrListUnref: attrList.        
	self cCode: 'g_object_unref(layout)'.
	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
	^ newW asOop: SmallInteger.
	! !

!RomePlugin methodsFor: 'primitives-text' stamp: 'bf 7/24/2006 15:42'!
primitiveShowZeroTerminatedUtf8String: aString x: x y: y
	"Show a string - ignores outline (use drawString primitive for outlines)"
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveShowZeroTerminatedUtf8StringXY' parameters: #(String Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].
	
	self cairoMoveTo: context x: x y: y.
	context cairoShowText: aString.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-drawing' stamp: 'bf 7/29/2006 00:17'!
primitiveStencilImage: formOop destLeft: dstL right: dstR top: dstT bottom: dstB srcLeft: srcL right: srcR top: srcT bottom: srcB
	| canvasOop context srcSurface |
	self var: 'context' type: 'cairo_t*'.
	self var: 'srcSurface' type: 'cairo_surface_t*'.

	canvasOop := self primitive: 'primitiveStencilImageSrcLRTBDestLRTB' parameters: #(Oop Float Float Float Float Float Float Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	srcSurface := self surfaceFrom: formOop.
	(srcSurface notNil and: [(self cairoSurfaceStatus: srcSurface) = 0]) ifFalse: [
		self fail: 'source surface invalid'. ^nil].

	(srcR ~= srcL and: [srcT ~= srcB]) ifTrue: [
		self cairoSave: context.
		self cairoTranslate: context x: dstL y: dstT.
		self cairoScale: context x: (dstR - dstL) / (srcR - srcL) y: (dstB - dstT) / (srcB - srcT).
		self cairoNewPath: context.
		self cairoRectangle: context x: 0.0 y: 0.0 w: srcR - srcL h: srcB - srcT.
		self cairoClip: context.
		self cairoMaskSurface: context surface: srcSurface x: 0.0-srcL y: 0.0-srcT.
		self cairoRestore: context].

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-transform' stamp: 'bf 7/24/2006 15:42'!
primitiveTransformBy: aTransform
	| canvasOop context m |
	self var: 'context' type: 'cairo_t*'.
	self var: 'm' type: 'cairo_matrix_t'.

	canvasOop := self primitive: 'primitiveTransformBy' parameters: #(FloatArray).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].
	
	self cairoMatrixInit: (self cCode: '&m' inSmalltalk: [m := aTransform])
		xx: (aTransform at: 0)
		yx: (aTransform at: 3)
		xy: (aTransform at: 1)
		yy: (aTransform at: 4)
		x0: (aTransform at: 2)
		y0: (aTransform at: 5).

	self cairoTransform: context matrix: (self cCode: '&m' inSmalltalk: [m]).

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'primitives-transform' stamp: 'bf 7/24/2006 15:42'!
primitiveTranslateByX: x y: y
	| canvasOop context |
	self var: 'context' type: 'cairo_t*'.

	canvasOop := self primitive: 'primitiveTranslateBy' parameters: #(Float Float).
	context := self contextFrom: canvasOop.
	interpreterProxy failed ifTrue: [^nil].

	self cairoTranslate: context x: x y: y.

	context cairoStatus ifTrue: [
		self fail: 'cairo error: %s' with: context cairoStatus cairoStatusToString. ^nil].
! !

!RomePlugin methodsFor: 'surfaces' stamp: 'bf 7/25/2006 18:30'!
registerSurface: surfaceHandle
	"Register the given surface, answer an ID or -1 on failure"
	| surfaceID |
	self var: 'surfaceHandle' type: 'cairo_surface_t*'.

	registerSurfaceFn isNil ifTrue: [
		self loadSurfacePlugin ifFalse: [^-1]].

	(self cCode: '(*registerSurfaceFn)((int)surfaceHandle, &surfaceDispatch, &surfaceID)'
		inSmalltalk: [surfaceID := surfaceHandle. true])
			ifFalse: [self fail: 'could not register surface'. ^-1].

	self debugCode: [self log: 'registered surface %p as %i' with: surfaceHandle with: surfaceID].

	surfaceID > maxSurfaceID
		ifTrue: [maxSurfaceID := surfaceID].

	^surfaceID! !

!RomePlugin methodsFor: 'contexts' stamp: 'bf 7/17/2006 19:39'!
setSource: context rgb: rgb alpha: alpha 

	| r g b |
	self returnTypeC: 'void'.
	self var: 'context' type: 'cairo_t*'.

	r := rgb >> 20 bitAnd: 1023.
	g := rgb >> 10 bitAnd: 1023.
	b := rgb >> 0 bitAnd: 1023.
	alpha = 255
		ifTrue: [self cairoSetSourceRgb: context r: r/1023.0 g: g /1023.0 b: b/1023.0]
		ifFalse: [self cairoSetSourceRgba: context r: r/1023.0 g: g /1023.0 b: b/1023.0 a: alpha/255.0].
! !

!RomePlugin methodsFor: 'surface-callbacks' stamp: 'bf 7/6/2006 14:10'!
showSurface: surfaceHandle x: x y: y w: w h: h
	self var: 'surfaceHandle' type: 'cairo_surface_t *'.
	^0! !

!RomePlugin methodsFor: 'initialize' stamp: 'yo 6/4/2007 13:13'!
shutdownModule
	| surface desc |
	self export: true.
	self var: 'surface' type: 'cairo_surface_t*'.
	self var: 'desc' type: 'PangoFontDescription*'.

	0 to: contexts lastIndex do: [:i |
		(contexts at: i) isNil ifFalse: [
			self log: 'context %i still active when unloading module!!' with: i.
			self cairoDestroy: (contexts at: i).
			contexts at: i put: nil]].

	0 to: maxSurfaceID do: [:id |
		surface := self findSurface: id.
		surface notNil ifTrue: [
			self log: 'surface %i still active when unloading module!!' with: id.
			self destroySurface: id]].
	
	0 to: 255 do: [:i |
		desc := fontDescriptions at: i.
		desc notNil ifTrue: [
			self pangoFontDescriptionFree: desc.
			fontDescriptions at: i put: nil]].
	
	defaultFontDescription notNil ifTrue: [
		self pangoFontDescriptionFree: defaultFontDescription.
		defaultFontDescription := nil.
	].

	^true! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/4/2007 12:06'!
sqCharCountIn: aString from: from to: to

	| index sqCount aChar |
	self var: 'aString' declareC: 'unsigned char* aString'.
	index := from.
	sqCount := 0.

	[index < to] whileTrue: [
		aChar := (aString at: index).
		aChar < 128 ifTrue: [
			index := index + 1.
		] ifFalse: [
			aChar < 2r11100000 ifTrue: [
				index := index + 2.
			] ifFalse: [
				aChar < 2r11110000 ifTrue: [
					index := index + 3.
				] ifFalse: [
					aChar < 2r11111000 ifTrue: [
						index := index + 4.
					] ifFalse: [
						aChar < 2r11111100 ifTrue: [
							index := index + 5.
						] ifFalse: [
							aChar >= 2r11111100 ifTrue: [
								index := index + 6.
							]]]]]].
		sqCount := sqCount + 1.
	].
	^ sqCount.

! !

!RomePlugin methodsFor: 'contexts' stamp: 'bf 7/24/2006 16:46'!
stroke: context from: canvasOop
	"stroke depending on canvasOop's stroke flag"

	| canvasFlags stroke rgb |
	self returnTypeC: 'void'.
	self var: 'context' type: 'cairo_t*'.

	canvasFlags := interpreterProxy fetchInteger: CanvasFlagsIndex ofObject: canvasOop.
	interpreterProxy failed
		ifTrue: [self fail: 'canvas flags not an integer']
		ifFalse: [
			stroke := canvasFlags bitAnd: CanvasFlagStroke.
			stroke ~= 0
				ifTrue: [
					rgb := interpreterProxy fetchInteger: CanvasStrokeColorIndex ofObject: canvasOop.
					self cairoSave: context.
					self setSource: context rgb: rgb alpha: stroke. 
					self cairoStroke: context.
					self cairoRestore: context]].! !

!RomePlugin methodsFor: 'surfaces' stamp: 'bf 7/7/2006 20:06'!
surfaceFrom: formOop
	"answer a surface by looking up its surface plugin ID stored as bits in formOop"
	| surfaceID |
	self returnTypeC: 'cairo_surface_t*'.
	self var: 'surface' type: 'cairo_surface_t*'.

	(interpreterProxy slotSizeOf: formOop) < FormInstSize ifTrue: [
		self fail: 'form oop invalid'. ^nil].

	surfaceID := interpreterProxy fetchInteger: FormBitsIndex ofObject: formOop.	
	interpreterProxy failed ifTrue: [
		self fail: 'form handle not an integer'. ^nil].
	
	^self findSurface: surfaceID.
! !

!RomePlugin methodsFor: 'primitives-pango' stamp: 'yo 6/11/2007 14:44'!
translateSqAttrsToPangoAttrs: sqAttrsArrayOop into: pangoAttrList
	| sqAttrsSize attrArray symbolOop symbol sqAttrsArray |
	self var: 'pangoAttrList' type: 'PangoAttrList *'.
	self var: 'sqAttrsArray' type: 'sqInt *'.
	self var: 'attrArray' type: 'sqInt'.
	self var: 'symbol' type: 'char *'.

	sqAttrsSize := interpreterProxy stSizeOf: sqAttrsArrayOop.
	sqAttrsArray := interpreterProxy firstIndexableField: sqAttrsArrayOop.
	self addDefaultInto: pangoAttrList.
	0 to: sqAttrsSize - 1 do: [:ind |
		attrArray := sqAttrsArray at: ind.
		symbolOop := interpreterProxy fetchPointer: 0 ofObject: attrArray.
		symbol := interpreterProxy firstIndexableField: symbolOop.
		(symbol at: 0) = 65 "A" ifTrue: [
			self addAlignment: attrArray into: pangoAttrList.
		].
		(symbol at: 0) = 67 "C" ifTrue: [
			self addColor: attrArray into: pangoAttrList.
		].
		(symbol at: 0) = 69 "E" ifTrue: [
			self addEmphasis: attrArray into: pangoAttrList.
		].
		(symbol at: 0) = 70 "F" ifTrue: [
			self addFont: attrArray into: pangoAttrList.
		].
		(symbol at: 0) = 76 "L" ifTrue: [
			self addLanguage: attrArray into: pangoAttrList.
		].
	].
	^ true.
! !

!RomePlugin methodsFor: 'surface-callbacks' stamp: 'bf 7/7/2006 18:08'!
unlockSurface: surfaceHandle x: x y: y w: w h: h
	self var: 'surfaceHandle' type: 'cairo_surface_t*'.

	self debugCode: [
		self log: 'unlock surface %p %i %i %i %i' with: surfaceHandle
			with: x with: y with: w with: h].

	(w > 0 and: [h > 0]) ifTrue: [
		self cairoSurfaceMarkDirtyRectangle: surfaceHandle x: x y: y w: w h: h].
	^true! !

!RomePlugin methodsFor: 'surfaces' stamp: 'bf 7/25/2006 18:30'!
unregisterSurface: surfaceID
	"Unregister the surface, answer true if successful"

	unregisterSurfaceFn isNil ifTrue: [
		self loadSurfacePlugin ifFalse: [^false]].

	self debugCode: [self log: 'unregistering surface %i' with: surfaceID].

	(self cCode: '(*unregisterSurfaceFn)(surfaceID)')
		ifFalse: [self fail: 'could not unregister surface %i' with: surfaceID. ^false].

	^true! !

!TSendNode methodsFor: '*Rome-Plugin' stamp: 'yo 12/11/2006 17:29'!
replaceCairoCalls
	((selector beginsWith: 'cairo') or: [selector beginsWith: 'pango']) ifTrue: [
		selector := RomePlugin camelToUnderscore: selector]! !
RomePlugin initialize!
